﻿using System;
using System.Text;
using System.IO;
using System.Reflection;
using ConsoleApplication1;


public class FixedLengthAttribute : Attribute
{
	public int Size {get; private set; }

	public FixedLengthAttribute(int size)
	{
		Size = size;
	}
}

public class ArrayLengthAttribute : Attribute
{
	public int Size { get; private set; }

	public ArrayLengthAttribute(int size)
	{
		Size = size;
	}
}

public class CannotDetermineFixedSizeForClassException : Exception
{
	private Type type;
	private string p;

	public CannotDetermineFixedSizeForClassException(Type type, string p)
	{
		this.type = type;
		this.p = p;
	}

}

class FixedSizedViolationException : Exception
{
	private Type type;
	private string p;

	public FixedSizedViolationException(Type type, string p)
	{
		// TODO: Complete member initialization
		this.type = type;
		this.p = p;
	}
}
/// <summary>
/// Sample implementaion for MemoryHelper
/// Thing to improve:
/// Caching of serializing mechanism
/// Multidimensional arrays
/// Handler strings shorter than class
/// </summary>
public static class MemoryHelper
{
	
	public static Encoding encoding = ASCIIEncoding.ASCII;

	public static String AsString(this object sourceRecord)
	{
		var stream = new MemoryStream();
		var writer = new BinaryWriter(stream);
		RecordToString(sourceRecord, writer);
		writer.Flush();
		var str = encoding.GetString(stream.ToArray());
		writer.Close();
		return str;
	}

	public static void SetFromString(this object targetRecord,string data)
	{
		var stream = new MemoryStream(encoding.GetBytes(data));
		var reader = new BinaryReader(stream);
		StringToRecord(targetRecord, reader);
	}

	static void RecordToString(object sourceRecord, System.IO.BinaryWriter writer)
	{
		foreach (var field in sourceRecord.GetType().GetFields())
		{
				FieldToString(field.GetValue(sourceRecord),field.FieldType, field, writer);
		}
	}

	static void FieldToString(object fieldValue,Type fieldType,FieldInfo field,System.IO.BinaryWriter writer)
	{
		if (field!=null && field.FieldType.IsArray)
		{
			Array array = (Array)fieldValue;
			var arrayElementType = field.FieldType.GetElementType();
			var attrs = field.GetCustomAttributes(typeof(ArrayLengthAttribute), false);
			int maxSize = attrs != null ? ((ArrayLengthAttribute)attrs[0]).Size : -1;
			int counter = 0;
			for (int i = array.GetLowerBound(0); i <= array.GetUpperBound(0); i++)
			{
				FieldToString(array.GetValue(i), arrayElementType, null, writer);
				counter++;
			}
			if (maxSize != -1 && counter != maxSize)

				throw new FixedSizedViolationException(field!=null?field.DeclaringType:null,field.Name);
		}		
		else if (fieldType == typeof(byte))
		{
			writer.Write((byte)fieldValue);
		}
		else if (fieldType == typeof(Int16))
		{
			writer.Write((Int16)fieldValue);
		}
		else if (fieldType == typeof(Int32))
		{
			writer.Write((Int32)fieldValue);
		}
		else if (fieldType == typeof(Int64))
		{
			writer.Write((Int64)fieldValue);
		}
		else if (fieldType == typeof(string))
		{
			object[] attrs = null;
			if (field!=null) attrs=field.GetCustomAttributes(typeof(FixedLengthAttribute), false);
			String strVal = (String)fieldValue;
			if (strVal == null) strVal = String.Empty;
			if (attrs != null)
			{
				int maxSize = ((FixedLengthAttribute)attrs[0]).Size;
				strVal = strVal.PadRight(maxSize);
				strVal = strVal.Substring(0, maxSize);
			}
			writer.Write(encoding.GetBytes(strVal));
		}
		else if (fieldType.IsClass)
		{
			if (fieldValue == null)
				writer.Write(encoding.GetBytes(String.Empty.PadLeft(GetClassLength(fieldType), (char)0)));
			else
				RecordToString(fieldValue, writer);
		}
	}

	private static int GetClassLength(Type type)
	{
		int totalLength = 0;
		foreach (var field in type.GetFields())
		{
			if (field.FieldType.IsArray)
			{
				var attrs = field.GetCustomAttributes(typeof(FixedLengthAttribute), false);
				if (attrs != null)
				{
					int maxSize = ((FixedLengthAttribute)attrs[0]).Size;
					totalLength += GetClassLength(field.FieldType) * maxSize;
				}
				else
					throw new CannotDetermineFixedSizeForClassException(type, field.Name);
			}
			else if (field.FieldType == typeof(byte))
			{
				totalLength += sizeof(byte);
			}
			else if (field.FieldType == typeof(Int16))
			{
				totalLength += sizeof(Int16);
			}
			else if (field.FieldType == typeof(Int32))
			{
				totalLength += sizeof(Int32);
			}
			else if (field.FieldType == typeof(Int64))
			{
				totalLength += sizeof(Int64);
			}
			else if (field.FieldType == typeof(string))
			{
				var attrs = field.GetCustomAttributes(typeof(FixedLengthAttribute), false);
				String strVal = String.Empty;
				if (attrs != null)
				{
					int maxSize = ((FixedLengthAttribute)attrs[0]).Size;
					totalLength += maxSize;
				}
				//else
				//	throw new CannotDetermineFixedSizeForClassException(type, field.Name);
			}
			else if (field.FieldType.IsClass)
			{
				totalLength += GetClassLength(field.FieldType);
			}
		}
		return totalLength;
	}


	static void StringToRecord(object targetRecord, System.IO.BinaryReader reader)
	{
		foreach (var field in targetRecord.GetType().GetFields())
		{
			if (field.FieldType == typeof(byte))
			{
				field.SetValue(targetRecord, reader.ReadByte());
			}
			else if (field.FieldType == typeof(Int16))
			{
				field.SetValue(targetRecord, reader.ReadInt16());
			}
			else if (field.FieldType == typeof(Int32))
			{
				field.SetValue(targetRecord, reader.ReadInt32());
			}
			else if (field.FieldType == typeof(Int64))
			{
				field.SetValue(targetRecord, reader.ReadInt64());
			}
			else if (field.FieldType == typeof(string))
			{
				var attrs = field.GetCustomAttributes(typeof(FixedLengthAttribute), false);
				var strVal = String.Empty;
				if (attrs != null)
				{
					int maxSize = ((FixedLengthAttribute)attrs[0]).Size;
					strVal = encoding.GetString(reader.ReadBytes(maxSize));
				}
				else
				{
					var memStream = (MemoryStream)reader.BaseStream;
					long bytesLeft = memStream.Length - memStream.Position;
					strVal = encoding.GetString(reader.ReadBytes((int)bytesLeft));
				}
				field.SetValue(targetRecord, strVal);
			}
			else if (field.FieldType.IsClass)
			{
				StringToRecord(field.GetValue(targetRecord), reader);
			}
		}
	}
}